﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Principal;
//using YOUR_IMPORTED_WEB_REFERENCE_FOR_SSRS_EXECUTION_SERVICE_HERE

namespace Org.SSRSPreview
{
    /// <summary>
    /// Export Formats
    /// </summary>
    public enum ExportFormat
    {
        /// <summary>XML</summary>
        XML,
        /// <summary>Comma Delimitted File
        CSV,
        /// <summary>TIFF image</summary>
        Image,
        /// <summary>PDF</summary>
        PDF,
        /// <summary>HTML (Web Archive)</summary>
        MHTML,
        /// <summary>HTML 4.0</summary>
        HTML4,
        /// <summary>HTML 3.2</summary>
        HTML32,
        /// <summary>Excel</summary>
        Excel,
        /// <summary>Word</summary>
        Word
    }

    public class ReportGenerator
    {
        /// <summary>
        /// Gets the string export format of the specified enum.
        /// </summary>
        /// <param name="Format">export format enum</param>
        /// <returns>enum equivalent string export format</returns>
        private string GetExportFormatString(ExportFormat Format)
        {
            switch (Format)
            {
                case ExportFormat.XML: return "XML";
                case ExportFormat.CSV: return "CSV";
                case ExportFormat.Image: return "IMAGE";
                case ExportFormat.PDF: return "PDF";
                case ExportFormat.MHTML: return "MHTML";
                case ExportFormat.HTML4: return "HTML4.0";
                case ExportFormat.HTML32: return "HTML3.2";
                case ExportFormat.Excel: return "EXCEL";
                case ExportFormat.Word: return "WORD";
                default:
                    return "PDF";
            }
        }

        /// <summary>
        /// generate a report
        /// </summary>
        /// <param name="ReportServer">report server</param>
        /// <param name="ReportServerExecutionURL">rendering execution url of report server</param>
        /// <param name="ReportServerUserName">user name to access report server</param>
        /// <param name="ReportServerPassword">password of the user name</param>
        /// <param name="ReportServerDomain">domain of the report server (for active directory)</param>
        /// <param name="ReportServerSubDir">folder of the report in the report server</param>
        /// <param name="FileFormat">output of the report</param>
        /// <param name="Parameters">parameters for the report</param>
        /// <param name="ReportContent">rendered report content</param>
        /// <param name="FileExtension">file extension of the report</param>
        /// <param name="MimeType">mime type of the generated file</param>
        public void GenerateReport(string ReportServer,
            string ReportServerExecutionURL,
            string ReportServerUserName,
            string ReportServerPassword,
            string ReportServerDomain,
            string ReportServerSubDir,
            string ReportFileName,
            ExportFormat FileFormat,
            Dictionary<string, string> Parameters,
            string DeviceInfo,
            out byte[] ReportContent,
            out string FileExtension,
            out string MimeType)
        {

            using (var reportExecutionService = new ReportExecutionService.ReportExecutionServiceSoapClient("ReportExecutionServiceSoap", ReportServerExecutionURL))
            {
                reportExecutionService.ClientCredentials.Windows.AllowedImpersonationLevel = TokenImpersonationLevel.Impersonation;
                reportExecutionService.ClientCredentials.UserName.UserName = ReportServerUserName;
                reportExecutionService.ClientCredentials.UserName.Password = ReportServerPassword;
                reportExecutionService.ClientCredentials.Windows.ClientCredential = new System.Net.NetworkCredential(ReportServerUserName, ReportServerPassword, ReportServerDomain);

                var parameters = Parameters.Select(p => new ParameterValue { Name = p.Key, Value = p.Value }).ToList();
                // Init Report to execute
                ServerInfoHeader serverInfoHeader;
                ExecutionInfo executionInfo;
                ExecutionHeader executionHeader = reportExecutionService.LoadReport(null, ReportFileName, null, out serverInfoHeader, out executionInfo);

                // Attach Report Parameters
                reportExecutionService.SetExecutionParameters(executionHeader, null, parameters.ToArray(), null, out executionInfo);

                // Render
                string encoding;
                Warning[] warnings;
                string[] streamIds;
                reportExecutionService.Render(executionHeader, null, GetExportFormatString(FileFormat), DeviceInfo, out ReportContent, out FileExtension, out MimeType, out encoding, out warnings, out streamIds);

            }
        }
    }
}
